/******************************************************************************
*******************************************************************************
  ECE 270 Fall 2016

  Kernel.c
	
  Christopher Theriault
  Homework #:5

  Copyright DigiPen (USA) Corporation
        All Rights Reserved
	
*******************************************************************************
******************************************************************************/

/******************************************************************************
*******************************************************************************
  References:

  [1] RM0090 Reference manual 
  STM32F405/415, STM32F407/417, STM32F427/437 and STM32F429/439 
    advanced ARM®-based 32-bit MCUs

  [2] UM1670 User manual
  Discovery kit with STM32F429ZI MCU

  [3] STM32F427xx/STM32F429xx Datasheet
  ARM Cortex-M4 32b MCU+FPU, 225DMIPS, up to 2MB Flash/256+4KB RAM, USB
    OTG HS/FS, Ethernet, 17 TIMs, 3 ADCs, 20 comm. interfaces, camera & LCD-TFT

  [4] PM0214 Programming manual
  STM32F3, STM32F4 and STM32L4 Series
    Cortex®-M4 programming manual

*******************************************************************************
******************************************************************************/

/*
        This code provides the Hardware Abstraction Layer (HAL) for the
    kernel.  This HAL only supports the STM32F429ZI microcontroller. 
*/

/******************************************************************************
*******************************************************************************
    Includes
*******************************************************************************
******************************************************************************/
#include 	"Kernel.h"

/******************************************************************************
*******************************************************************************
    Definitions
*******************************************************************************
******************************************************************************/
// Definitions required by the kernel
#define		STACK_SIZE		0x040
#define		MAX_TASKS		0x03
#define		READY_LIST		(OSp_GetList(0x00))

// Optional definitions not required to produce working code
#ifndef			MAX_WAIT
	#define		BON(X)			|=(X)
	#define		BOFF(X)			&=~(X)
	#define		BTOG(X)			^=(X)
	#define		MAX_WAIT		0xFFFF
#endif		//	MAX_WAIT

/******************************************************************************
*******************************************************************************
    Declarations & Types
*******************************************************************************
******************************************************************************/
typedef	void (* OS_TaskAddress)(void);	

typedef enum {READY=OS_READY, RUNNING=OS_RUNNING} taskState_t;

typedef struct {
    unsigned int	*sp;
    unsigned int	taskID;
    taskState_t		taskState;
    OS_TaskAddress	taskAddress;
    unsigned int	basePriority;
    unsigned int	curPriority;
    unsigned int        execution;
    unsigned int        deadline;
    unsigned int	stack[OS_STACK_SIZE];
} TCB;	

typedef		struct {
    unsigned int        listSize;
    TCB			*TCBPtr[MAX_TASKS];
}listObj;

// Globals used by the system
TCB             Tasks[OS_MAX_TASKS];
TCB             *OS_TaskNEW;
TCB             *OS_TaskRUNNING;
kernelErrors	OS_Error;
unsigned char 	newTaskID;

/******************************************************************************
*******************************************************************************
    Prototypes
*******************************************************************************
******************************************************************************/
void
OS_IdleTask(void);

void
OSp_SetError(kernelErrors error);

listObj *
OSp_GetList(unsigned int);

unsigned int
OSp_ListPush(listObj *, TCB *);

TCB *
OSp_ListPop (listObj *);

void
OSp_ListPriority (listObj *);
/******************************************************************************
*******************************************************************************
    Helper Functions
*******************************************************************************
******************************************************************************/

/******************************************************************************
    OS_InitKernel
		
      Prepares the Kernel for use, but does not start any services.  No OS_
    function should be called until after this one has executed.
******************************************************************************/    
unsigned int    
OS_InitKernel(unsigned int numTasks, unsigned int stackSize) {
    unsigned int i;
    listObj *listREADY = READY_LIST;

    if ((numTasks < 0x01) || (numTasks > (MAX_TASKS - 0x01))) {
        OSp_SetError(TASK_LIMIT);
	goto OS_InitKernelExit;
    } // end if

    if ((stackSize < 0x01) || (stackSize > STACK_SIZE) || 
                              (stackSize % 0x04 != 0x00)) {
	OSp_SetError(STACK_LIMIT);
	goto OS_InitKernelExit;
    } // end if

    OS_InitKernelHAL();

    // Set software values	
   // newTaskID = 0x00;
    listREADY->listSize = 0x00;
    for (i = 0x00; i < MAX_TASKS; i++){
        listREADY->TCBPtr[i] = 0x00;
    } // end for

    OS_CreateTask(OS_IdleTask, 0x00, 0x01, 0x00);
    OS_TaskNEW = &Tasks[0x00];
    OS_TaskRUNNING = &Tasks[0x00];
    
    return NO_ERROR;
OS_InitKernelExit:
    return 0x00;
} // end OS_InitKernel

/******************************************************************************
    OS_TaskCreate
		
      Takes the assigned function pointer and uses it to create a kernel task
    that is ready for execution.
******************************************************************************/
unsigned int
OS_CreateTask(void (* newTask)(void),   unsigned int prio, 
                                        unsigned int execution, 
                                        unsigned int deadline) {
    static unsigned int newTaskID = 0x00;

    Tasks[newTaskID].taskAddress = newTask;
    Tasks[newTaskID].taskID = newTaskID;
    Tasks[newTaskID].taskState = READY;
    Tasks[newTaskID].basePriority = prio;
    Tasks[newTaskID].curPriority = prio;
    Tasks[newTaskID].execution = execution;
    Tasks[newTaskID].deadline = deadline;

    // This line sets the PC area of the stack
    Tasks[newTaskID].stack[OS_STACK_SIZE - 0x01] = 0xDEADBEEF;
    // STATUS register (Thumb bit is set)
    Tasks[newTaskID].stack[OS_STACK_SIZE - 0x02] = 0x01000000;
    // This line sets the PC area of the stack
    Tasks[newTaskID].stack[OS_STACK_SIZE - 0x03] = (unsigned int)newTask; 	
    // R14
    Tasks[newTaskID].stack[OS_STACK_SIZE - 0x04] = 0x14141414;
    // R12
    Tasks[newTaskID].stack[OS_STACK_SIZE - 0x05] = 0x12121212;
    // R3
    Tasks[newTaskID].stack[OS_STACK_SIZE - 0x06] = 0x03030303;
    // R2
    Tasks[newTaskID].stack[OS_STACK_SIZE - 0x07] = 0x02020202;
    // R1
    Tasks[newTaskID].stack[OS_STACK_SIZE - 0x08] = 0x01010101;
    // R0
    Tasks[newTaskID].stack[OS_STACK_SIZE - 0x09] = 0x00000000;
    // R11
    Tasks[newTaskID].stack[OS_STACK_SIZE - 0x0A] = 0x11111111;
    // R10
    Tasks[newTaskID].stack[OS_STACK_SIZE - 0x0B] = 0x10101010;
    // R9
    Tasks[newTaskID].stack[OS_STACK_SIZE - 0x0C] = 0x09090909;
    // R8
    Tasks[newTaskID].stack[OS_STACK_SIZE - 0x0D] = 0x08080808;
    // R7
    Tasks[newTaskID].stack[OS_STACK_SIZE - 0x0E] = 0x07070707;
    // R6
    Tasks[newTaskID].stack[OS_STACK_SIZE - 0x0F] = 0x06060606;
    // R5
    Tasks[newTaskID].stack[OS_STACK_SIZE - 0x10] = 0x05050505;
    // R4
    Tasks[newTaskID].stack[OS_STACK_SIZE - 0x11] = 0x04040404;
    // Set the stack pointer to the end of the stack (pos 0x11)
    Tasks[newTaskID].sp = 
        (unsigned int *)&Tasks[newTaskID].stack[OS_STACK_SIZE - 0x11];
    // Move to the next available ID
    ++newTaskID;

    return NO_ERROR;
} // end OS_TaskCreate

/******************************************************************************
    OS_IdleTask
		
      This task should always be created, have the lowest possible (worst) 
    priority, and never be prevented from running.
******************************************************************************/
void
OS_IdleTask(void) {
    while (0x01) {
        ;
    } // end while

} // end OS_IdleTask

/******************************************************************************
    OSp_ScheduleTask
		
      In this version of the code, simply alternate between .
******************************************************************************/
void
OSp_ScheduleTask(void) {
    listObj *listREADY = READY_LIST;
	
    // Get the highest priority task
    OS_TaskNEW = OSp_ListPop(listREADY);
    // Increase priority of every task that didn't get to run (idle is excl.)
    OSp_ListPriority(listREADY);
    // Reset the priority of the task that just ran and place into queue
    OS_TaskRUNNING->curPriority = OS_TaskRUNNING->basePriority;
    OSp_ListPush(listREADY, OS_TaskRUNNING);

} // end OSp_ScheduleTask

/******************************************************************************
    OS_GetError
		
      Returns the current error code being reported by the kernel.
******************************************************************************/
kernelErrors
OS_GetError(void) {
	
    return OS_Error;
} // end OS_GetError

/******************************************************************************
    OSp_SetError
		
      Sets the current error code being reported by the kernel.
******************************************************************************/
void
OSp_SetError(kernelErrors error) {
	
    OS_Error = error;
} // end OS_SetError

/******************************************************************************
    OSp_GetList
		
	Returns the head pointer of the selected linked list:
			[0] = The READY list
******************************************************************************/
listObj *
OSp_GetList (unsigned int list) {
/*
    static listObj listREADY;
	
    switch (list) {
	case 0x00:
            return &listREADY;
        default:
            return 0x00;
    } // end switch
    */
    return 0x00;
} // end OSp_GetList

/******************************************************************************
    OSp_ListPush
		
        Inserts the given task into the given task list based on the 
    current priority level of the task.  It is implicitly assumed that 
    higher priority values are better than lower ones and that only the
    system idle task will have priority zero.
******************************************************************************/
unsigned int
OSp_ListPush (listObj *list, TCB *newEntry) {
    unsigned int insert;
    unsigned int i;
	
    // Check the error conditions
    if (!list || !newEntry) {
	goto OSp_ListInsertExit;
    } // end if
    if (list->listSize == MAX_TASKS) {
	goto OSp_ListInsertExit;
    } // end if
	
    // Special case: list is empty
    if (!list->listSize) {
	list->TCBPtr[0x00] = newEntry;
    } // end if
    else { // List is not empty
	insert = list->listSize;
		
        // Count back to find the insertion point
        while (insert && (list->TCBPtr[insert - 0x01]->curPriority >= 
                                                      newEntry->curPriority)) {
            insert--;
        } // end while
		
        // Move down the end items in the list if needed
        if (insert != list->listSize) {
            for (i = list->listSize; i > insert; i--) {
                list->TCBPtr[i] = list->TCBPtr[i - 0x01];
            } // end for
        } // end if
        list->TCBPtr[insert] = newEntry;
    } // end else
	
    list->listSize++;
	
    return NO_ERROR;	
OSp_ListInsertExit:	
    OSp_SetError(LIST_ERROR);
    return 0x00;
} // end OSp_ListInsert

/******************************************************************************
    OSp_ListPop
		
      Returns the highest priority task which should be at the end of the list.
******************************************************************************/
TCB *
OSp_ListPop (listObj * list) {
    TCB *pop = 0x00;
	
    if (list) {
        if (list->listSize) {
            pop = list->TCBPtr[list->listSize - 0x01];
            list->TCBPtr[list->listSize - 0x01] = 0x00;
            list->listSize--;
        } // end if
    } // end if
		
    return pop;
} // end OSp_ListPop
 
/******************************************************************************
    OSp_ListPriority
		
      Returns the highest priority task which should be at the end of the list.
******************************************************************************/
void
OSp_ListPriority (listObj * list) {
    unsigned int i;
	
    // Only act if we have a valid list
    if (list) {
	// Only change priority if we have more than the idle task
	if (list->listSize > 0x01) {
            for (i = 0x01; i < list->listSize; i++) {
                if (list->TCBPtr[i]->curPriority < OS_MAX_PRIORITY) {
                    list->TCBPtr[i]->curPriority++;
		} // end if
            } // end for
	} // end if
    } // end if
	
} // end OSp_ListPriority

// EOF    Kernel.c
// Note: Some IDEs generate warnings if a file doesn't end in whitespace,
//  but Embedded Studio doesn't seem to be one of them.

